from projectreport.analyzer.parsers.javascript import PackageJSONParser
from projectreport.analyzer.parsers.python.init import PythonInitParser
from projectreport.analyzer.parsers.python.setup_cfg import PythonSetupCfgParser
from projectreport.analyzer.parsers.python.setup_py import PythonSetupPyParser

PARSER_EXTENSIONS = {"py": PythonInitParser}

PARSER_DOC_FILES = {
    "__init__.py": PythonInitParser,
    "setup.py": PythonSetupPyParser,
    "setup.cfg": PythonSetupCfgParser,
    "package.json": PackageJSONParser,
}



import abc
from typing import ClassVar

from cached_property import cached_property

from projectreport.analyzer.parsers.data_types import ParserDataType


class Parser(abc.ABC):
    data_type: ClassVar[ParserDataType]

    def __init__(self, path: str):
        self.path = path

    @cached_property
    def docstring(self):
        raise NotImplementedError

    @cached_property
    def version(self):
        raise NotImplementedError



import abc
from pathlib import Path

from cached_property import cached_property

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.data_types import ParserDataType


class SingleFileParser(Parser, abc.ABC):
    data_type = ParserDataType.SINGLE_FILE

    @cached_property
    def contents(self) -> str:
        with open(self.path, encoding="utf8") as f:
            file_contents = f.read()
        return file_contents

    @cached_property
    def file_name(self) -> str:
        return Path(self.path).with_suffix("").name

    @classmethod
    def matches_path(cls, path: str) -> bool:
        raise NotImplementedError



from os import getenv
from typing import Any, Dict, Optional, Sequence, Tuple, TypedDict

import github.GithubException
from cached_property import cached_property
from github import Github

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.data_types import ParserDataType
from projectreport.analyzer.parsers.url import URLParser
from projectreport.tools.monkey_patch_github import (
    monkey_patch_github_obj_for_throttling,
)
from projectreport.version import Version

access_token = getenv("GITHUB_TOKEN")

gh = Github(access_token)
monkey_patch_github_obj_for_throttling(gh)


class GithubData(TypedDict):
    name: str
    description: Optional[str]
    version: Optional[str]


class GithubParser(URLParser):
    def __init__(self, path: str):
        self.repo_key = _github_url_to_owner_and_name(path)
        self.github_repo = gh.get_repo(self.repo_key)
        monkey_patch_github_obj_for_throttling(self.github_repo)
        super().__init__(path)

    @staticmethod
    def find_github_url(urls: Sequence[str]) -> Optional[str]:
        for url in urls:
            if "github.com" in url:
                return url
        return None

    @cached_property
    def contents(self) -> str:
        raise NotImplementedError

    @cached_property
    def file_name(self) -> str:
        return f"repo:github:{self.github_repo.name}"

    @cached_property
    def parsed(self) -> GithubData:
        data = GithubData(
            name=self.github_repo.name,
            description=self.github_repo.description,
            version=None,
        )
        try:
            release = self.github_repo.get_latest_release()
        except github.UnknownObjectException:
            pass
        else:
            if release is not None:
                data["version"] = release.tag_name
        return data

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None
        return self.parsed.get("description")

    @cached_property
    def version(self) -> Optional[Version]:
        if self.parsed is None:
            return None
        version_str = self.parsed.get("version")
        if version_str is None:
            return None
        return Version.from_str(version_str)

    @classmethod
    def matches_path(cls, path: str) -> bool:
        return "github.com" in path


def _github_url_to_owner_and_name(url: str) -> str:
    if "github.com" not in url:
        raise ValueError(f"got a non-Github URL {url}")
    return "/".join(url.split("/")[-2:])



import abc
from pathlib import Path
from typing import Sequence

from cached_property import cached_property

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.data_types import ParserDataType


class FolderParser(Parser, abc.ABC):
    """
    :param path: This should be the path of a folder, rather than a path to a file
      that the singular parsers accept.
    :param file_names: A list of file names that should be analyzed within the folder.
    """

    data_type = ParserDataType.FOLDER

    def __init__(
        self,
        path: str,
        file_names: Sequence[str],
    ):
        self.file_names = file_names
        super().__init__(path)

    @classmethod
    def matches_path(cls, path: str, file_names: Sequence[str]) -> bool:
        raise NotImplementedError



from typing import Final, List, Optional, Sequence, Type

from cached_property import cached_property
from typing_extensions import TypeGuard

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.data_types import ParserDataType
from projectreport.analyzer.parsers.folder import FolderParser
from projectreport.analyzer.parsers.github import GithubParser
from projectreport.analyzer.parsers.multi.file import MultiFileParser
from projectreport.analyzer.parsers.url import URLParser
from projectreport.logger import logger
from projectreport.parser_types import StandaloneParser
from projectreport.version import Version

# TODO: Improve user ability to customize parsers. Currently they need to mutate the PARSERS
# array and PARSER_DOC_FILES dict to customize parsers as the matches_path classmethod does
# not take into account custom parsers passed by constructor.
PARSERS: Final[List[Type[StandaloneParser]]] = [
    MultiFileParser,
    GithubParser,
]


class MainMultiParser(Parser):
    """
    The main parser that uses other parsers, both single- and multi- together to
    determine information about a folder.
    """

    def __init__(
        self,
        path: str,
        file_names: Sequence[str],
        urls: Optional[Sequence[str]] = None,
        parsers: Optional[List[Type[StandaloneParser]]] = None,
    ):
        """
        :param path: This should be the path of a folder, rather than a path to a file
          that the singular parsers accept.
        :param parsers: Defaults to PARSERS.
        """
        self.parsers = parsers or PARSERS
        self.file_names = file_names
        self.urls = urls or []
        super().__init__(path)

    @classmethod
    def matches_path(
        cls, path: str, file_names: Sequence[str], urls: Optional[Sequence[str]] = None
    ) -> bool:
        urls = urls or []
        for parser in PARSERS:
            if _parser_matches_path(parser, path, file_names, urls):
                return True
        return False

    @cached_property
    def docstring(self) -> Optional[str]:
        return self._get_attr_from_first_parser_to_return_non_none("docstring")

    @cached_property
    def version(self) -> Optional[Version]:
        return self._get_attr_from_first_parser_to_return_non_none("version")

    def _get_attr_from_first_parser_to_return_non_none(self, attr: str):
        for parser in self.parsers:
            if _parser_matches_path(parser, self.path, self.file_names, self.urls):
                parser_obj = _construct_parser(
                    parser, self.path, self.file_names, self.urls
                )
                if parser_obj is None:
                    continue
                value = getattr(parser_obj, attr)
                if value is not None:
                    return value
        return None


def _parser_matches_path(
    parser: Type[StandaloneParser],
    folder: str,
    file_names: Sequence[str],
    urls: Sequence[str],
) -> bool:
    if _is_folder_parser(parser):
        return parser.matches_path(folder, file_names)
    elif _is_url_parser(parser):
        return any([parser.matches_path(url) for url in urls])
    raise NotImplementedError(
        f"No handling for {parser} with data type {parser.data_type}"
    )


def _construct_parser(
    parser: Type[StandaloneParser],
    folder: str,
    file_names: Sequence[str],
    urls: Sequence[str],
) -> Optional[StandaloneParser]:
    if _is_folder_parser(parser):
        return parser(folder, file_names)
    elif _is_url_parser(parser):
        for url in urls:
            if parser.matches_path(url):
                return parser(url)
        logger.warn(f"The URL parser {parser} did not match the given urls {urls}")
        return None
    raise NotImplementedError(
        f"No handling for {parser} with data type {parser.data_type}"
    )


def _is_folder_parser(parser: Type[StandaloneParser]) -> TypeGuard[Type[FolderParser]]:
    return parser.data_type == ParserDataType.FOLDER


def _is_url_parser(parser: Type[StandaloneParser]) -> TypeGuard[Type[URLParser]]:
    return parser.data_type == ParserDataType.URL



from pathlib import Path
from typing import Final, List

GITHUB_REPO_URL = "https://github.com/nickderobertis/github-project-example"
TESTS_DIR = Path(__file__).parent
PROJECT_DIR = TESTS_DIR.parent
TEST_FILES_BASE_PATH = TESTS_DIR / "input_data"

PYTHON_EXAMPLES_FOLDER = TEST_FILES_BASE_PATH / "python_examples"
PYTHON_DUNDER_VERSION_PROJECT_NAME = "dunder_version"
PYTHON_DUNDER_VERSION_PROJECT_PATH = (
    PYTHON_EXAMPLES_FOLDER / PYTHON_DUNDER_VERSION_PROJECT_NAME
)
PYTHON_SETUP_PY_VERSION_PROJECT_NAME = "setup_py"
PYTHON_SETUP_PY_VERSION_PROJECT_PATH = (
    PYTHON_EXAMPLES_FOLDER / PYTHON_SETUP_PY_VERSION_PROJECT_NAME
)
PYTHON_SETUP_CFG_VERSION_PROJECT_NAME = "setup_cfg"
PYTHON_SETUP_CFG_VERSION_PROJECT_PATH = (
    PYTHON_EXAMPLES_FOLDER / PYTHON_SETUP_CFG_VERSION_PROJECT_NAME
)
PYTHON_PROJECT_PATHS: Final[List[str]] = sorted(
    [
        str(PYTHON_DUNDER_VERSION_PROJECT_PATH),
        str(PYTHON_SETUP_PY_VERSION_PROJECT_PATH),
        str(PYTHON_SETUP_CFG_VERSION_PROJECT_PATH),
    ]
)

JS_PROJECT_NAME = "js_example"
JS_PROJECT_PATH = TEST_FILES_BASE_PATH / JS_PROJECT_NAME

GIT_PROJECT_NAME = "git_example"

REPORTS_NAME = "reports"
REPORTS_FOLDER = TEST_FILES_BASE_PATH / REPORTS_NAME



from pathlib import Path

from projectreport.analyzer.parsers.github import GithubParser
from projectreport.analyzer.parsers.multi.file import MultiFileParser
from projectreport.analyzer.parsers.multi.main import MainMultiParser
from projectreport.analyzer.parsers.python.init import PythonInitParser
from projectreport.analyzer.parsers.python.setup_cfg import PythonSetupCfgParser
from projectreport.analyzer.parsers.python.setup_py import PythonSetupPyParser
from projectreport.version import Version
from tests.config import (
    GITHUB_REPO_URL,
    PYTHON_DUNDER_VERSION_PROJECT_PATH,
    PYTHON_SETUP_CFG_VERSION_PROJECT_PATH,
    PYTHON_SETUP_PY_VERSION_PROJECT_PATH,
)
from tests.fixtures.temp_dir import temp_folder


def test_init_py_parser():
    input_file = PYTHON_DUNDER_VERSION_PROJECT_PATH / "__init__.py"
    assert PythonInitParser.matches_path(input_file)
    parser = PythonInitParser(input_file)
    assert parser.docstring == "An example Python package for testing purposes"
    assert parser.version == Version.from_str("0.0.1")


def test_setup_py_parser():
    input_file = PYTHON_SETUP_PY_VERSION_PROJECT_PATH / "setup.py"
    assert PythonSetupPyParser.matches_path(input_file)
    parser = PythonSetupPyParser(input_file)
    assert (
        parser.docstring
        == "[from setup.py] An example Python package for testing purposes (version from setup.py)"
    )
    assert parser.version == Version.from_str("0.0.1")


def test_setup_cfg_parser():
    input_file = PYTHON_SETUP_CFG_VERSION_PROJECT_PATH / "setup.cfg"
    assert PythonSetupCfgParser.matches_path(input_file)
    parser = PythonSetupCfgParser(input_file)
    assert (
        parser.docstring
        == "[from setup.cfg] An example Python package for testing purposes (version from setup.cfg)"
    )
    assert parser.version == Version.from_str("1.0.0")


def test_github_parser():
    url = GITHUB_REPO_URL
    assert GithubParser.matches_path(url)
    parser = GithubParser(url)
    assert parser.docstring == "Example Github Project for project-report tests"
    assert parser.version == Version.from_str("1.0.0")


def test_multi_file_parser():
    input_folder = PYTHON_SETUP_CFG_VERSION_PROJECT_PATH
    input_files = ["__init__.py", "setup.cfg"]
    assert MultiFileParser.matches_path(input_folder, input_files)
    parser = MultiFileParser(input_folder, input_files)
    assert (
        parser.docstring
        == "[from __init__.py] An example Python package for testing purposes (version from setup.cfg)"
    )
    assert parser.version == Version.from_str("1.0.0")


# TODO: additional tests for main multi parser


def test_main_multi_parser_on_folder():
    input_folder = PYTHON_SETUP_CFG_VERSION_PROJECT_PATH
    input_files = ["__init__.py", "setup.cfg"]
    assert MainMultiParser.matches_path(input_folder, input_files)
    parser = MainMultiParser(input_folder, input_files)
    assert (
        parser.docstring
        == "[from __init__.py] An example Python package for testing purposes (version from setup.cfg)"
    )
    assert parser.version == Version.from_str("1.0.0")


def test_main_multi_parser_on_urls(temp_folder: Path):
    input_folder = str(temp_folder)
    input_files = []
    urls = [GITHUB_REPO_URL]
    assert MainMultiParser.matches_path(input_folder, input_files, urls)
    parser = MainMultiParser(input_folder, input_files, urls)
    assert parser.docstring == "Example Github Project for project-report tests"
    assert parser.version == Version.from_str("1.0.0")



import ast
import warnings

from cached_property import cached_property

from projectreport.analyzer.parsers.file import SingleFileParser


class PythonParser(SingleFileParser):
    @cached_property
    def parsed(self):
        try:
            return ast.parse(self.contents)
        except SyntaxError:
            warnings.warn(f"Could not parse {self.path} due to SyntaxError")
            return None



import abc

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.data_types import ParserDataType


class URLParser(Parser, abc.ABC):
    data_type = ParserDataType.URL

    @classmethod
    def matches_path(cls, path: str) -> bool:
        raise NotImplementedError



import ast
from configparser import ConfigParser, ParsingError
from pathlib import Path
from typing import Optional

from cached_property import cached_property

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.python.base import PythonParser
from projectreport.version import Version

# NOTE: This does not use the PythonParser as a base class because this is not a Python file
# being parsed (it is an ini file with the cfg extension)


class PythonSetupCfgParser(Parser):
    @cached_property
    def parsed(self) -> Optional[ConfigParser]:
        # Try to load the file as an ini file using configparser
        try:
            config = ConfigParser()
            config.read(self.path)
            return config
        except ParsingError:
            return None

    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None

        # Return the description value from the metadata section, if it exists
        return _get_string_from_config(self.parsed, "metadata", "description")

    @cached_property
    def version(self) -> Optional[Version]:
        if self.parsed is None:
            return None

        # Return the version value from the metadata section, if it exists
        version_str = _get_string_from_config(self.parsed, "metadata", "version")
        if version_str is None:
            return None
        return Version.from_str(version_str)

    @classmethod
    def matches_path(cls, path: str) -> bool:
        return Path(path).name == "setup.cfg"


def _get_string_from_config(
    config: ConfigParser, section: str, option: str
) -> Optional[str]:
    # Try to get the value from the config
    try:
        str_with_quotes = config.get(section, option)
    except (KeyError, AttributeError):
        return None
    else:
        # Remove the quotes from the string, but only if they exist
        if str_with_quotes.startswith('"') and str_with_quotes.endswith('"'):
            return str_with_quotes[1:-1]
        elif str_with_quotes.startswith("'") and str_with_quotes.endswith("'"):
            return str_with_quotes[1:-1]
        else:
            return str_with_quotes
